package whilelang

import java.io.{InputStream, IOException}

sealed abstract class Token
case object EOF extends Token           { override def toString = "<EOF>" }
case object WHILE extends Token         { override def toString = "while" }
case object IF extends Token            { override def toString = "if" }
case object ELSE extends Token          { override def toString = "else" }
case object PRINTLN extends Token       { override def toString = "println" }
case object ASSUME extends Token        { override def toString = "assume" }
case object ASSERT extends Token        { override def toString = "assert" }
case object GETINT extends Token        { override def toString = "getInt" }
case object SEMICOLON extends Token     { override def toString = ";" }
case object LPAREN extends Token        { override def toString = "(" }
case object RPAREN extends Token        { override def toString = ")" }
case object EQSIGN extends Token        { override def toString = "=" }
case object LBRACE extends Token        { override def toString = "{" }
case object RBRACE extends Token        { override def toString = "}" }
case object BANG extends Token          { override def toString = "!" }
case object PLUS extends Token          { override def toString = "+" }
case object MINUS extends Token         { override def toString = "-" }
case object TIMES extends Token         { override def toString = "*" }
case object DIV extends Token           { override def toString = "/" }
case object PERCENT extends Token       { override def toString = "%" }
case object LT extends Token            { override def toString = "<" }
case object GT extends Token            { override def toString = ">" }
case object GEQ extends Token           { override def toString = ">=" }
case object LEQ extends Token           { override def toString = "<=" }
case object EQUALS extends Token        { override def toString = "==" }
case object NOTEQUALS extends Token     { override def toString = "!=" }
case object AND extends Token           { override def toString = "&&" }
case object OR extends Token            { override def toString = "||" }
case object COMMA extends Token         { override def toString = "," }
case object BOX extends Token           { override def toString = "[]" }
case class NUMLIT(value: Int) extends Token { override def toString = value.toString }
case class ID(value: String) extends Token { override def toString = value }
case class STRLIT(value: String) extends Token { override def toString = "\""+value+"\"" }

class Lexer {  
  val keywords = Map[String,Token](
    ("if" -> IF),
    ("else" -> ELSE),
    ("while" -> WHILE),
    ("println" -> PRINTLN),
    ("assume" -> ASSUME),
    ("assert" -> ASSERT),
    ("getInt" -> GETINT))
  
  // the input stream
  private var in: InputStream = _
  // the last char seen in the input stream
  private var currentChar: Char = _
  // the current token type under consideration (useless if not an identifier or a literal)
  private var currentToken: Token = _
  // a buffer to store textual information for parametric tokens
  private val buffer: StringBuffer = new StringBuffer
  
  /** Sets the input stream to use for lexing. */
  def setInputStream(input: InputStream): Unit = { in = input; nextChar }
  
  /** What will mark the end of the input stream. */
  private val EndOfFile: Char = java.lang.Character.MAX_VALUE
  
  /** Used to detect \r\n pairs and ouput only \n for them. */
  private var previousChar: Char = _
  
  /** Puts the next character in the input stream in currentChar, or the special character EndOfFile if the stream is exhausted. */
  private def nextChar: Unit = {
    def readChar: Char = in.read match {
      case c if c < 0 => EndOfFile
      case c => c.asInstanceOf[Char]
    }

    currentChar match {                                                                                                                                                      
      case EndOfFile => return
      case _ => ;                                                                                                                 
    }
    
    try {                                                                                                                                                           
      currentChar  = readChar
      previousChar = if ((previousChar == '\r') && (currentChar == '\n')) readChar else currentChar                                                                                                   
      currentChar = if (previousChar == '\r') '\n' else previousChar                                                                                                                       
    } catch {                                                                                                                                                         
      case e: IOException => Error(e.getMessage)                                                                                              
    }
  }
  
  /** Gets rid of whitespaces and comments and calls readToken to get the next token. */
  def nextToken: Token = {
    while(currentChar == '/' || Character.isWhitespace(currentChar)) {
      if(currentChar == '/') {
        nextChar
        
        if(currentChar == '/') {
          // skips end-of-line comments
          while(currentChar != '\n' && currentChar != EndOfFile) nextChar
        } else if(currentChar == '*') {
          // skips block comments
          var foundEnd: Boolean = false
          while(!foundEnd) {
            while(currentChar != '*') {
              if(currentChar == EndOfFile) Error("Unterminated block comment")
              nextChar
            }
            nextChar
            if(currentChar == '/') { foundEnd = true; nextChar }
          }
        } else {
          return DIV
        }
      } else {
        nextChar
      }
    }
    
    return readToken
  }
  
  /** Reads the next token from the stream. */
  private def readToken: Token = {    
    currentChar match {
      case EndOfFile => EOF
    
      case _ if Character.isLetter(currentChar) => {
        buffer.setLength(0)
        do {
          buffer.append(currentChar)
          nextChar
        } while (Character.isLetterOrDigit(currentChar) || currentChar == '_');
        val str: String = buffer.toString
        keywords.get(str) match {
          case None => ID(str)
          case Some(token) => token
        }
      }
    
      case '0' => { nextChar; NUMLIT(0) }
      
      case _ if Character.isDigit(currentChar) => {
        buffer.setLength(0)
        do {
          buffer.append(currentChar)
          nextChar
        } while (Character.isDigit(currentChar))
        val num: Int = Integer.parseInt(buffer.toString)
        NUMLIT(num)
      }
    
      case '"' => {
        buffer.setLength(0)
        nextChar
        while (currentChar != '"') {
          if (currentChar == '\n' || currentChar == EndOfFile) {
            Error("Unterminated string")
          }
          buffer.append(currentChar)
          nextChar
        }
        nextChar
        val str: String = buffer.toString
        STRLIT(str)
      }
    
      case ';' => { nextChar; SEMICOLON }
      case ',' => { nextChar; COMMA }
      case '!' => { nextChar; if(currentChar == '=') { nextChar; NOTEQUALS } else { BANG } }
      case '(' => { nextChar; LPAREN }
      case ')' => { nextChar; RPAREN }
      case '{' => { nextChar; LBRACE }
      case '}' => { nextChar; RBRACE }
      case '<' => { nextChar; if(currentChar == '=') { nextChar; LEQ } else { LT } }
      case '>' => { nextChar; if(currentChar == '=') { nextChar; GEQ } else { GT } }
      case '+' => { nextChar; PLUS }
      case '-' => { nextChar; MINUS }
      case '*' => { nextChar; TIMES }
      case '%' => { nextChar; PERCENT }
      case '=' => { nextChar; if(currentChar == '=') { nextChar; EQUALS } else { EQSIGN } }
      case '&' => { nextChar; if(currentChar == '&') { nextChar; AND } else { Error("Single '&'") } }
      case '|' => { nextChar; if(currentChar == '|') { nextChar; OR } else { Error("Single '|'") } }
      case '[' => { nextChar; if(currentChar == ']') { nextChar; BOX } else { Error("'[' withouth following ']'") } }
    
      case _ => { Error("Invalid character: " + currentChar) }
    }
  }
}
