package whilelang

import java.io.{InputStream,IOException}

/** LL(1) parser for the while language grammar. */
class Parser extends Lexer {  
  def parseInputStream(in: InputStream): Command = {
    setInputStream(in)
    readToken 
    parseGoal
  }
  
  /** Store the current token, and one lookahead token, as read from the lexer. */
  private var currentToken: Token = _
  
  def readToken: Unit = {
    currentToken = nextToken
  }
  
  val anyId = ID("Identifier")
  val anyNumLit = new NUMLIT(0) { override def toString = "number literal" }
  val anyStrLit = STRLIT("String literal")
  
  /** ''Eats'' the expected token, or terminates with an error. */
  private def eat(token: Token): Unit = (currentToken, token) match {
    case (ID(_), ID(_)) => readToken
    case (NUMLIT(_), NUMLIT(_)) => readToken
    case (STRLIT(_), STRLIT(_)) => readToken
    case (a, b) if a == b => readToken
    case _ => expected(token)
  }
    
  /** Complains that what was found was not expected. The method accepts
   * arbitrarily many arguments of type TokenClass */
  private def expected(token: Token, more: Token*): Nothing = {
    Error("Expected: " + (token::more.toList).mkString(" or ") + ", found: " + currentToken)
  }
    
  private def parseGoal: Command = {
    val coms = parseCommands
    eat(EOF)
    Block(coms)
  }
  
  private def parseCommand: Command = currentToken match {
    case PRINTLN => {
      readToken
      eat(LPAREN)
      val printTxt = parseStringLit
      eat(COMMA)
      val printId = parseIdentifier
      eat(RPAREN)
      eat(SEMICOLON)
      Print(printTxt, printId)
    }
    case ID(_) => {
      val assignId = parseIdentifier
      eat(EQSIGN)

      currentToken match {
        case GETINT => {
          readToken
          eat(LPAREN)
          val lb = parseInt
          eat(COMMA)
          val ub = parseInt
          if(ub < lb) {
            Error("Upper bound smaller than lower bound in getInt(" + lb + ", " + ub + ");")
          }
          eat(RPAREN)
          eat(SEMICOLON)
          GetInt(assignId, lb, ub)
        }

        case _ => {
          val expr = parseTerm
          eat(SEMICOLON)
          Assignment(assignId, expr)
        }
      }
    }
    case IF => {
      readToken
      eat(LPAREN)
      val cond = parseBoolExpr
      eat(RPAREN)
      val then = parseCommand
      currentToken match {
        case ELSE => {
          readToken
          val elze = parseCommand
          IfThenElse(cond, then, elze)
        }
        case _ => IfThenElse(cond, then, Skip())
      }
    }
    case WHILE => {
      readToken
      eat(LPAREN)
      val cond = parseBoolExpr
      eat(RPAREN)
      val com = parseCommand
      WhileLoop(cond, com)
    }

    case ASSUME => {
      readToken
      eat(LPAREN)
      val cond = parseBoolExpr
      eat(RPAREN)
      eat(SEMICOLON)
      Assume(cond)
    }

    case ASSERT => {
      readToken
      eat(LPAREN)
      val cond = parseBoolExpr
      eat(RPAREN)
      eat(SEMICOLON)
      Assert(cond)
    }

    case LBRACE => {
      readToken
      val ret = Block(parseCommands)
      eat(RBRACE)
      ret
    }

    case LPAREN => {
      readToken
      val c1 = parseCommand
      eat(BOX)
      val c2 = parseCommand
      eat(RPAREN)
      Choice(c1, c2)
    }

    case _ => expected(PRINTLN, anyId, IF, WHILE, LBRACE, ASSUME, ASSERT, LPAREN)
  }
  
  private def parseCommands: List[Command] = {
    var lst: List[Command] = Nil
    while(isFirstOfCommand) {
      lst = parseCommand :: lst
    }
    lst.reverse
  }
  
  private def isFirstOfCommand: Boolean = currentToken match {
    case ID(_) | IF | WHILE | PRINTLN | LBRACE | ASSUME | ASSERT | LPAREN => true
    case _ => false
  }
  
  private def parseStringLit: String = currentToken match {
    case STRLIT(value) => { readToken; value }
    case _ => expected(anyStrLit)
  }
  
  private def parseIdentifier: String = currentToken match {
    case ID(value) => { readToken; value }
    case _ => expected(anyId)
  }
  
  // all expressions
  private def parseBoolExpr: BoolExpr = parseExpr7
  
  // expr7 ::= expr6 ('||' expr6)*
  private def parseExpr7: BoolExpr = {
    var e6 = parseExpr6
    while(currentToken == OR) {
      readToken
      e6 = Or(e6, parseExpr6)
    }
    e6
  }
  
  // expr6 ::= expr5 ('&&' expr5)*
  private def parseExpr6: BoolExpr = {
    var e5 = parseExpr5
    while(currentToken == AND) {
      readToken
      e5 = And(e5, parseExpr5)
    }
    e5
  }
  
  // expr5 ::= expr4 (('>' | '<' | '==') expr4)*
  private def parseExpr5: BoolExpr = {
    if(currentToken == BANG) {
      readToken
      Negation(parseExpr5)
    } else if(currentToken == LPAREN) {
      readToken
      val ret = parseBoolExpr
      eat(RPAREN)
      ret
    } else {
      var e4 = parseTerm
      if(currentToken == EQUALS) {
        readToken
        Equal(e4, parseTerm)
      } else if(currentToken == NOTEQUALS) {
        readToken
        Negation(Equal(e4, parseTerm))
      } else if(currentToken == LT) {
        readToken
        LessThan(e4, parseTerm)
      } else if(currentToken == GT) {
        readToken
        GreaterThan(e4, parseTerm)
      } else if(currentToken == LEQ) {
        readToken
        Negation(LessThan(parseTerm, e4))
      } else if(currentToken == GEQ) {
        readToken
        Negation(GreaterThan(parseTerm, e4))
      } else {
        expected(EQUALS, LT, GT, LEQ, GEQ)
      }
    }
  }
  
  private def parseTerm = parseExpr4

  // expr4 ::= expr3 (('+' | '-') expr3)*
  private def parseExpr4: Term = {
    var e3 = parseExpr3
    while(currentToken == PLUS || currentToken == MINUS) {
      if(currentToken == PLUS) {
        readToken
        e3 = Plus(e3, parseExpr3)
      } else {
        readToken
        e3 = Minus(e3, parseExpr3)
      }
    }
    e3
  }
  
  // expr3 ::= expr2a (('*' | '/' | '%') expr2a)*
  private def parseExpr3: Term = {
    var e2 = parseExpr2a
    while(currentToken == DIV || currentToken == TIMES || currentToken == PERCENT) {
      if(currentToken == DIV) {
        readToken
        val e2b = parseExpr2a
        e2 = e2b match {
          case Constant(c) => Divide(e2, c)
          case _ => Error("Division by a non-constant term.")
        }
      } else if(currentToken == TIMES) {
        readToken
        e2 = e2 match {
          case Constant(c) => Times(c, parseExpr2a)
          case _ => Error("Multiplication by a non-constant term.")
        }
      } else {
        readToken
        val e2b = parseExpr2a
        e2 = e2b match {
          case Constant(c) => Modulo(e2, c)
          case _ => Error("Modulo with a non-constant base.")
        }
      }
    }
    e2
  }
  
  // expr2a ::= !expr2a | expr2b
  private def parseExpr2a: Term = parseExpr2b
  private def parseExpr2b = parseExpr1
  
  private def parseExpr1: Term = currentToken match {
      case NUMLIT(value) => { val ret = Constant(value); readToken; ret }
      case LPAREN => { readToken; val res = parseTerm; eat(RPAREN); res }
      case ID(value) => { readToken; Variable(value) }      
      case _ => expected(anyNumLit, LPAREN, anyId)
  }

  private def parseInt: Int = currentToken match {
    case NUMLIT(value) => { readToken; value }
    case _ => expected(anyNumLit)
  }
}
